from __future__ import annotations

from typing import TYPE_CHECKING

from tach import errors
from tach import filesystem as fs
from tach.check import check
from tach.filesystem import get_project_config_path
from tach.parsing import dump_project_config_to_yaml

if TYPE_CHECKING:
    from pathlib import Path

    from tach.core import ProjectConfig


def sync_dependency_constraints(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str] | None = None,
) -> ProjectConfig:
    """
    Update project configuration with auto-detected dependency constraints.
    This is additive, meaning it will create dependencies to resolve existing errors,
    but will not remove any constraints.
    """
    check_result = check(
        project_root=project_root,
        project_config=project_config,
        exclude_paths=exclude_paths,
    )
    for error in check_result.errors:
        error_info = error.error_info
        if error_info.is_dependency_error:
            project_config.add_dependency_to_module(
                error_info.source_module, error_info.invalid_module
            )

    return project_config


def prune_dependency_constraints(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str] | None = None,
) -> ProjectConfig:
    """
    Build a minimal project configuration with auto-detected module dependencies.
    """
    # Force module dependencies to be empty so that we can figure out the minimal set
    project_config = project_config.model_copy(
        update={
            "modules": [
                module.model_copy(update={"depends_on": []})
                for module in project_config.modules
            ]
        }
    )

    sync_dependency_constraints(
        project_root=project_root,
        project_config=project_config,
        exclude_paths=exclude_paths,
    )

    return project_config


def sync_project(
    project_root: Path,
    project_config: ProjectConfig,
    add: bool = False,
    exclude_paths: list[str] | None = None,
) -> None:
    tach_yml_path = get_project_config_path(project_root)
    if tach_yml_path is None:
        raise errors.TachError(
            "Unexpected error. Could not find configuration file during 'sync'."
        )

    if add:
        project_config = sync_dependency_constraints(
            project_root=project_root,
            project_config=project_config,
            exclude_paths=exclude_paths,
        )
    else:
        project_config = prune_dependency_constraints(
            project_root=project_root,
            project_config=project_config,
            exclude_paths=exclude_paths,
        )

    tach_yml_content = dump_project_config_to_yaml(project_config)
    fs.write_file(str(tach_yml_path), tach_yml_content)


__all__ = ["sync_project", "prune_dependency_constraints"]



from __future__ import annotations

from itertools import chain
from pathlib import Path
from unittest.mock import patch

import pytest

from tach.check import check_import
from tach.cli import tach_check
from tach.core import (
    ModuleConfig,
    ModuleNode,
    ModuleTree,
)
from tach.core.config import RootModuleConfig
from tach.filesystem import validate_project_modules


@pytest.fixture
def example_dir() -> Path:
    current_dir = Path(__file__).parent
    return current_dir / "example"


@pytest.fixture
def test_config() -> ModuleConfig:
    return ModuleConfig(path="test", strict=False)


@pytest.fixture
def module_tree() -> ModuleTree:
    return ModuleTree(
        root=ModuleNode(
            is_end_of_path=False,
            full_path="",
            config=None,
            children={
                "domain_one": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_one",
                    config=ModuleConfig(
                        path="domain_one",
                        depends_on=["domain_one.subdomain", "domain_three"],
                        strict=True,
                    ),
                    interface_members=["public_fn"],
                    children={
                        "subdomain": ModuleNode(
                            is_end_of_path=True,
                            full_path="domain_one.subdomain",
                            config=ModuleConfig(
                                path="domain_one.subdomain", strict=True
                            ),
                            children={},
                        )
                    },
                ),
                "domain_two": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_two",
                    config=ModuleConfig(
                        path="domain_two", depends_on=["domain_one"], strict=False
                    ),
                    children={
                        "subdomain": ModuleNode(
                            is_end_of_path=True,
                            full_path="domain_two.subdomain",
                            config=ModuleConfig(
                                path="domain_two",
                                depends_on=["domain_one"],
                                strict=False,
                            ),
                            children={},
                        )
                    },
                ),
                "domain_three": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_three",
                    config=ModuleConfig(path="domain_three", strict=False),
                    children={},
                ),
            },
        )
    )


@pytest.mark.parametrize(
    "valid_modules,invalid_modules",
    [
        (["mod.a"], []),
        ([], ["mod.b"]),
        (["mod.a", "mod.b"], ["mod.c"]),
        (["mod.a", "mod.b"], ["mod.c", "mod.d"]),
    ],
)
def test_validate_project_modules(tmp_path, valid_modules, invalid_modules):
    def mock_fs_check(source_root, module_path):
        return module_path in valid_modules

    mock_source_root = tmp_path / "src"
    with patch("tach.filesystem.module_to_pyfile_or_dir_path", wraps=mock_fs_check):
        result = validate_project_modules(
            mock_source_root,
            [ModuleConfig(path=path) for path in chain(valid_modules, invalid_modules)],
        )
        assert set(mod.path for mod in result.valid_modules) == set(valid_modules)
        assert set(mod.path for mod in result.invalid_modules) == set(invalid_modules)


@patch("tach.filesystem.module_to_pyfile_or_dir_path")
def test_validate_project_modules_root_is_always_valid(tmp_path):
    result = validate_project_modules(tmp_path / "src", [RootModuleConfig()])
    assert (
        len(result.valid_modules) == 1 and result.valid_modules[0] == RootModuleConfig()
    )
    assert not result.invalid_modules


@pytest.mark.parametrize(
    "file_mod_path,import_mod_path,expected_result",
    [
        ("domain_one", "domain_one", True),
        ("domain_one", "domain_one.subdomain", True),
        ("domain_one", "domain_one.core", True),
        ("domain_one", "domain_three", True),
        ("domain_two", "domain_one", True),
        ("domain_two", "domain_one.public_fn", True),
        ("domain_two.subdomain", "domain_one", True),
        ("domain_two", "external", True),
        ("external", "external", True),
        ("domain_two", "domain_one.private_fn", False),
        ("domain_three", "domain_one", False),
        ("domain_two", "domain_one.core", False),
        ("domain_two.subdomain", "domain_one.core", False),
        ("domain_two", "domain_three", False),
        ("domain_two", "domain_two.subdomain", False),
        ("external", "domain_three", False),
    ],
)
def test_check_import(module_tree, file_mod_path, import_mod_path, expected_result):
    check_error = check_import(
        module_tree=module_tree,
        file_mod_path=file_mod_path,
        import_mod_path=import_mod_path,
    )
    result = check_error is None
    assert result == expected_result


def test_valid_example_dir(example_dir):
    project_root = example_dir / "valid"
    with pytest.raises(SystemExit) as exc_info:
        tach_check(project_root=project_root)
    assert exc_info.value.code == 0


def test_valid_example_dir_monorepo(example_dir):
    project_root = example_dir / "monorepo"
    with pytest.raises(SystemExit) as exc_info:
        tach_check(project_root=project_root)
    assert exc_info.value.code == 0



from __future__ import annotations

import os
import tempfile
from pathlib import Path

import pytest

from tach.extension import get_project_imports


# Utility function to create temporary files with content
def create_temp_file(directory, filename, content):
    filepath = os.path.join(directory, filename)
    with open(filepath, "w") as f:
        f.write(content)
    return filepath


@pytest.fixture
def temp_project():
    with tempfile.TemporaryDirectory() as project_root:
        project_root = Path(project_root)
        # Creating some sample Python files in a nested structure
        (project_root / "a" / "b").mkdir(parents=True, exist_ok=True)
        (project_root / "d").mkdir(parents=True, exist_ok=True)
        (project_root / "local" / "g").mkdir(parents=True, exist_ok=True)
        (project_root / "parent").mkdir(parents=True, exist_ok=True)

        file1_content = """
import os
from local.file2 import b
"""
        file2_content = """
from ..file1 import y
"""
        file3_content = """
if TYPE_CHECKING:
    from local.file2 import c
"""
        file4_content = """
# tach-ignore
from a.b import c
# tach-ignore d.e.f
from d.e import f

import file3
"""

        create_temp_file(project_root, "file1.py", file1_content)
        create_temp_file(project_root, "local/file2.py", file2_content)
        create_temp_file(project_root, "file3.py", file3_content)
        create_temp_file(project_root, "file4.py", file4_content)

        yield project_root


def test_regular_imports(temp_project):
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "file1.py"),
        ignore_type_checking_imports=True,
    )
    expected = [("local.file2.b", 3)]
    assert result == expected


def test_relative_imports(temp_project):
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "local/file2.py"),
        ignore_type_checking_imports=True,
    )
    expected = [("file1.y", 2)]
    assert result == expected


def test_ignore_type_checking_imports(temp_project):
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "file3.py"),
        ignore_type_checking_imports=True,
    )
    expected = []
    assert result == expected


def test_include_type_checking_imports(temp_project):
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "file3.py"),
        ignore_type_checking_imports=False,
    )
    expected = [("local.file2.c", 3)]
    assert result == expected


def test_mixed_imports(temp_project):
    mixed_content = """
import sys
if TYPE_CHECKING:
    from .file2 import c
from ..file1 import x
"""
    create_temp_file(temp_project, "local/file4.py", mixed_content)
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "local/file4.py"),
        ignore_type_checking_imports=True,
    )
    expected = [("file1.x", 5)]
    assert result == expected

    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "local/file4.py"),
        ignore_type_checking_imports=False,
    )
    expected = [("local.file2.c", 4), ("file1.x", 5)]
    assert result == expected


def test_external_imports(temp_project):
    external_content = """
import os
from external_module import something
"""
    create_temp_file(temp_project, "file5.py", external_content)
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "file5.py"),
        ignore_type_checking_imports=True,
    )
    expected = []  # 'os' and 'external_module' are not within the project root
    assert result == expected


def test_external_and_internal_imports(temp_project):
    mixed_content = """
import os
from file1 import c
from external_module import something
"""
    create_temp_file(temp_project, "file6.py", mixed_content)
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "file6.py"),
        ignore_type_checking_imports=True,
    )
    expected = [
        ("file1.c", 3),
    ]
    assert result == expected


def test_ignored_imports(temp_project):
    result = get_project_imports(
        str(temp_project),
        ".",
        str(temp_project / "file4.py"),
        ignore_type_checking_imports=True,
    )
    expected = [("file3", 7)]
    assert result == expected


def test_file_outside_source_root(temp_project, tmp_path):
    mixed_content = """
import os
from file1 import c
from external_module import something
"""

    path_outside_source_root = tmp_path / "outside_src_root.py"
    path_outside_source_root.write_text(mixed_content)

    result = get_project_imports(
        str(temp_project),
        ".",
        str(path_outside_source_root),
        ignore_type_checking_imports=True,
    )
    expected = [
        ("file1.c", 3),
    ]
    assert result == expected
